from dataclasses import dataclass


@dataclass
class Resources:
    # Format to: secret_key
    default_config = """\
# Flask-BigApp Config File
#
# Updates the Flask app config with the variables below.
# If any variable below does not exist in the standard Flask env vars it is created and will be accessible using
# app.config["YOUR_VAR_NAME"]. All key names defined below will be capitalised when imported.
#
[FLASK]
APP_NAME = "app"
VERSION = "0.0.1"
SECRET_KEY = "change_me"
DEBUG = false
TESTING = false
SESSION_TIME = 480
ERROR_404_HELP = true
SQLALCHEMY_TRACK_MODIFICATIONS = false
EXPLAIN_TEMPLATE_LOADING = false


[DATABASE]
#
# [database.main] is loaded as SQLALCHEMY_DATABASE_URI
# type = mysql / postgresql / sqlite
# if type = sqlite, config parser will ignore username and password
#
[DATABASE.MAIN]
ENABLED = true
TYPE = "sqlite"
DATABASE_NAME = "database"
LOCATION = "db"
PORT = ""
USERNAME = "user"
PASSWORD = "password"

# Adding another database is as simple as adding a new section
# [DATABASE.ANOTHER] will then be accessible using SQLALCHEMY_BINDS
# The bind key will be stored as a lowercase value, so "ANOTHER" will
# be accessible as "another"
# You can then use the bind key in the model as follows:
# class MyModel(db.Model):
#     __bind_key__ = "another"
#     ...
#
#[DATABASE.ANOTHER]
#ENABLED = true
#TYPE = "sqlite"
#DATABASE_NAME = "database_another"
#LOCATION = "db"
#PORT = ""
#USERNAME = "user"
#PASSWORD = "password"
#
# You can add as many databases as you want.\
"""
    # Format to: static_url_path
    default_theme_config = """\
enabled = "yes"
static_folder = "static"
static_url_path = "{static_url_path}"\
"""



import typing as t
from functools import wraps

from flask import flash
from flask import redirect
from flask import session
from flask import url_for


def login_check(
        bool_session_key: str,
        endpoint_redirect: str,
        redirect_on_value: bool = False,
        endpoint_kwargs: t.Optional[t.Dict[str, t.Union[str, int]]] = None,
        message: t.Optional[str] = None,
        message_category: str = "message"
):
    """
    A decorator that checks if the specified session key exists, redirects to the specified endpoint if it does not
    exist or match the specified value.

    Example of a route that requires a user to be logged in:

    @bp.route("/admin", methods=["GET"])
    @login_check('logged_in', 'blueprint.login_page', message="Login needed")
    def admin_page():
        ...

    Example of a route that if the user is already logged in, redirects to the specified endpoint:

    @bp.route("/login-page", methods=["GET"])
    @login_check('logged_in', 'blueprint.admin_page', redirect_on_value=True, message="Already logged in")
    def login_page():
        ...

    :param bool_session_key: The session key to check for.
    :param endpoint_redirect: The endpoint to redirect to
                              if the session key does not exist or
                              match the specified value.
    :param redirect_on_value: The boolean value to check the session key against.
    :param endpoint_kwargs: A dictionary of keyword arguments to pass to the redirect endpoint.
    :param message: If a message is specified, a flash message is shown.
    :param message_category: The category of the flash message.
    """

    def login_check_wrapper(func):
        @wraps(func)
        def inner(*args, **kwargs):
            if bool_session_key not in session:
                if message:
                    flash(message, message_category)
                if endpoint_kwargs is None:
                    return redirect(url_for(endpoint_redirect))
                return redirect(url_for(endpoint_redirect, **endpoint_kwargs))

            if session[bool_session_key] == redirect_on_value:
                if message:
                    flash(message, message_category)
                if endpoint_kwargs is None:
                    return redirect(url_for(endpoint_redirect))
                return redirect(url_for(endpoint_redirect, **endpoint_kwargs))

            return func(*args, **kwargs)

        return inner

    return login_check_wrapper


def permission_check(
        iter_session_key: str,
        endpoint_redirect: str,
        redirect_if_no_match: t.Optional[list] = None,
        endpoint_kwargs: t.Optional[t.Dict[str, t.Union[str, int]]] = None,
        message: t.Optional[str] = None,
        message_category: str = "message"
):
    """
    A decorator that checks if the specified session key exists, redirects to the specified endpoint if
    the session key does not exist or if the session key does not contain the specified values.

    Example:

    @bp.route("/admin-page", methods=["GET"])
    @login_check('logged_in', 'blueprint.login_page') <- can be mixed with login_check
    @permission_check('permissions', 'www.index', redirect_if_no_match=['admin'],
                      message="You don't have the correct permissions to view this page.")
    def admin_page():
        ...

    IF redirect_if_no_match is None, the default value of ['admin'] is set.

    :param iter_session_key: The session key to check for.
    :param endpoint_redirect: The endpoint to redirect to if the
                              session key does not exist or does not contain the
                              specified values.
    :param redirect_if_no_match: A list of values that the session key must contain.
    :param endpoint_kwargs: A dictionary of keyword arguments to pass to the redirect endpoint.
    :param message: If a message is specified, a flash message is shown.
    :param message_category: The category of the flash message.
    """

    if redirect_if_no_match is None:
        redirect_if_no_match = ['admin']

    def permission_check_wrapper(func):
        @wraps(func)
        def inner(*args, **kwargs):
            if iter_session_key not in session:
                if message:
                    flash(message, message_category)
                if endpoint_kwargs is None:
                    return redirect(url_for(endpoint_redirect))
                return redirect(url_for(endpoint_redirect, **endpoint_kwargs))

            for item in redirect_if_no_match:
                if item not in session[iter_session_key]:
                    if message:
                        flash(message, message_category)
                    if endpoint_kwargs is None:
                        return redirect(url_for(endpoint_redirect))
                    return redirect(url_for(endpoint_redirect, **endpoint_kwargs))

            return func(*args, **kwargs)

        return inner

    return permission_check_wrapper


class BigAppSecurity:
    """
    A class that provides route function decorators to check for session variables and their values.
    """

    @staticmethod
    def login_required(
            redirect_endpoint: str,
            bool_key_name: str,
            show_message: bool = True,
            message: str = "You need to be logged in to access this page."
    ):

        def login_required_wrapper(func):
            @wraps(func)
            def secure_function(*args, **kwargs):
                if bool_key_name not in session:
                    if show_message:
                        flash(message)
                    return redirect(url_for(redirect_endpoint))

                if not session[bool_key_name]:
                    if show_message:
                        flash(message)
                    return redirect(url_for(redirect_endpoint))

                return func(*args, **kwargs)

            return secure_function

        return login_required_wrapper

    @staticmethod
    def no_login_required(
            redirect_endpoint: str,
            session_bool_key_name: str,
            show_message: bool = True,
            message: str = "You are already logged in."
    ):

        def no_login_required_wrapper(func):
            @wraps(func)
            def secure_function(*args, **kwargs):
                if session_bool_key_name in session:
                    if session[session_bool_key_name]:
                        if show_message:
                            flash(message)
                        return redirect(url_for(redirect_endpoint))

                return func(*args, **kwargs)

            return secure_function

        return no_login_required_wrapper

    @staticmethod
    def permission_required(
            redirect_endpoint: str,
            session_list_key: str,
            permission_needed: str,
            show_message: bool = True,
            message: str = "You don't have the required permissions to access this page."
    ):

        def permission_required(func):
            @wraps(func)
            def secure_function(*args, **kwargs):
                if session_list_key not in session:
                    if show_message:
                        flash(message)
                    return redirect(url_for(redirect_endpoint))

                if permission_needed not in session[session_list_key]:
                    if show_message:
                        flash(message)
                    return redirect(url_for(redirect_endpoint))

                return func(*args, **kwargs)

            return secure_function

        return permission_required



def loader(app):
    from markupsafe import Markup

    @app.template_filter('example__hello_world')
    def example__filter_(value: str) -> str:
        """
        Usage

        {{ "World" | example__hello_world }} => Hello World, again!
        """
        return Markup(f"Hello {value}, again!")

    @app.template_filter('example__num_to_month')
    def example__num_to_month(num: str) -> str:
        """
        Usage

        {{ 1 | example__num_to_month }} => January
        """
        if isinstance(num, int):
            num = str(num)

        months = {
            "1": "January", "2": "February", "3": "March", "4": "April", "5": "May", "6": "June",
            "7": "July", "8": "August", "9": "September", "10": "October", "11": "November", "12": "December",
        }

        num = num.replace("0", "")
        if num in months:
            return months[num]
        return "Month not found"



import logging
import os
from importlib import import_module
from inspect import getmembers
from inspect import isclass
from pathlib import Path
from typing import Dict, Union, Optional, Any

from flask import Blueprint
from flask import Flask
from flask import session
from flask_sqlalchemy.model import DefaultMeta
from toml import load as toml_load

from .blueprint import BigAppBlueprint
from .helpers import init_app_config
from .registeries import ModelRegistry
from .resources import Resources
from .utilities import cast_to_bool, cast_to_import_str, deprecated


class BigApp(object):
    _app: Flask
    _app_name: str
    _app_path: Path
    _app_folder: Path

    __model_registry__: ModelRegistry

    config_path: Path
    config: Dict
    themes: Dict[str, Path]

    def __init__(
            self,
            app: Optional[Flask] = None,
            app_config_file: Optional[str] = None
    ) -> None:

        if app is not None:
            self.init_app(app, app_config_file)

    def init_app(
            self,
            app: Flask,
            app_config_file: Optional[str] = os.environ.get("BA_CONFIG")
    ) -> None:

        """
        Initializes the application.

        -> Expects a Flask application

        Can be passed an SQLAlchemy object, doing this will enable the use of the model_class.

        Optional config from_file passed in. If no config from_file is given an
        attempt will be made to read from the environment for the variable BA_CONFIG
        If the environment variable is not found it will create a new config from_file
        called default.config.toml and proceed to load the values from there.
        """

        if app is None:
            raise ImportError(
                "No app was passed in, do ba = BigApp(flaskapp) or app.init_app(flaskapp)")
        if not isinstance(app, Flask):
            raise TypeError(
                "The app that was passed in is not an instance of Flask")
        if app_config_file is None:
            app_config_file = "default.config.toml"

        self._app = app
        self._app_name = app.name
        self._app_path = Path(self._app.root_path)
        self._app_folder = self._app_path.parent

        self.__model_registry__ = ModelRegistry()

        self.config_path = self._app_path / app_config_file
        self.config = init_app_config(self.config_path, self._app)
        self.themes = dict()

    def init_session(self) -> None:
        """
        Initialize the session variables found in the config from_file.
        Use this method in the before_request route.
        """
        if self.config.get("SESSION"):
            for key, value in self.config.get("SESSION", {}).items():
                if key not in session:
                    session[key] = value

    def import_builtins(self, folder: Union[str, Path] = "builtins") -> None:
        """
        Folder must be relative ( from_folder="here" not from_folder="/home/user/app/from_folder" )
        """
        if isinstance(folder, str):
            builtins_folder = Path(self._app_path / folder)
        else:
            builtins_folder = folder

        if builtins_folder.is_dir():
            with self._app.app_context():
                for py_file in builtins_folder.glob("*.py"):
                    module = import_module(
                        f"{cast_to_import_str(self._app_name, builtins_folder)}.{py_file.stem}")
                    if hasattr(module, "loader"):
                        module.loader(self._app)

    def import_blueprints(self, folder: str) -> None:
        """
        Imports all the blueprints in the given from_folder.

        Folder must be relative ( from_folder="here" not from_folder="/home/user/app/from_folder" )
        """

        folder_path = Path(self._app_path / folder)

        for potential_bp in folder_path.iterdir():
            self.import_blueprint(potential_bp)

    def import_blueprint(self, blueprint: Union[str, Path]):
        """
        Imports a single blueprint from the given path.

        Path must be relative ( path="here" not path="/home/user/app/from_folder" )
        """
        if isinstance(blueprint, str):
            potential_bp = Path(self._app_path / blueprint)
        else:
            potential_bp = blueprint

        if potential_bp.is_dir():
            try:
                module = import_module(cast_to_import_str(self._app_name, potential_bp))
                for dir_item in dir(module):
                    _ = getattr(module, dir_item)
                    if isinstance(_, BigAppBlueprint):
                        if _.enabled:
                            self._app.register_blueprint(_)
                        break
            except Exception as e:
                logging.critical(f"{e}",
                                 f"Error importing blueprint: from {potential_bp}")

    def import_themes(self, themes_folder: str) -> None:
        """
        Folder must be relative ( from_folder="here" not from_folder="/home/user/app/from_folder" )
        """
        folder_path = Path(self._app_path / themes_folder)
        for theme in folder_path.iterdir():
            self.import_theme(theme)

    def import_theme(self, theme_folder: Union[str, Path]):
        """
        Folder must be relative ( from_folder="here" not from_folder="/home/user/app/from_folder" )
        """
        if isinstance(theme_folder, str):
            theme = Path(self._app_path / theme_folder)
        else:
            theme = theme_folder

        if theme.is_dir():
            if theme.name in self.themes:
                raise ImportError(
                    f"Theme {theme.name} is already registered, theme folders must be unique\n"
                    f"Importing from {theme}"
                )

            static_folder = Path(theme / "static")
            template_folder = Path(theme / "templates")
            nested_template_folder = Path(theme / "templates" / theme.name)
            builtins_folder = Path(theme / "builtins")

            config = Path(theme / "config.toml")

            if not static_folder.exists():
                logging.debug(
                    f"Static from_folder for {theme.name} was not found, skipping")

            if not template_folder.exists():
                raise NotADirectoryError(
                    f"Template from_folder for {theme.name} was not found")

            if not nested_template_folder.exists():
                raise NotADirectoryError(
                    f"Nested template from_folder for {theme.name} was not found, "
                    f"\nshould look like: {nested_template_folder}")

            if not config.exists():
                logging.debug(
                    f"Config from_file for {theme.name} was not found, creating default")
                config.write_text(Resources.default_theme_config.format(
                    static_url_path=f"/{theme.name}/static"))

            loaded_config = toml_load(config)

            theme_bp = Blueprint(
                name=theme.name,
                import_name=f"{theme.name}",
                static_folder=f"{theme.absolute()}/static",
                template_folder=f"{theme.absolute()}/templates",
                static_url_path=f"{loaded_config.get('static_url_path', f'/{theme.name}/static')}",
            )

            if cast_to_bool(loaded_config.get("enabled")):

                if builtins_folder.exists():
                    self.import_builtins(builtins_folder)

                self._app.register_blueprint(theme_bp)
                self.themes.update({theme.name: theme.absolute()})

    def import_models(
            self,
            from_file: Optional[Union[str, Path]] = None,
            from_folder: Optional[Union[str, Path]] = None,
    ) -> None:
        """
        Imports model files from a single from_file or a from_folder. Both are allowed to be set.

        File and Folder must be relative ( from_folder="here" not from_folder="/home/user/app/from_folder" )
        """

        def model_processor(path: Path):
            """
            Picks apart the model from_file and builds a registry of the models found.
            """
            import_string = cast_to_import_str(self._app_name, path)
            try:
                model_module = import_module(import_string)
                for model_object_members in getmembers(model_module, isclass):
                    if import_string in model_object_members[1].__module__:
                        name = model_object_members[0]
                        model = model_object_members[1]
                        if not hasattr(model, "__tablename__"):
                            raise AttributeError(
                                f"{name} is not a valid model")

                        self.__model_registry__.add(name, model)

            except ImportError as e:
                logging.critical("Error importing model: ", e, f" {import_string}")

        if from_file is None and from_folder is None:
            raise ImportError(
                "No model from_file or from_folder was passed in")

        if from_file is not None:
            if isinstance(from_file, Path):
                file_path = from_file
            else:
                file_path = Path(self._app_path / from_file)

            if file_path.is_file() and file_path.suffix == ".py":
                model_processor(file_path)
            else:
                from_folder = from_file

        if from_folder is not None:
            if isinstance(from_folder, Path):
                folder_path = from_folder
            else:
                folder_path = Path(self._app_path / from_folder)

            if folder_path.is_dir():
                for model_file in folder_path.glob("*.py"):
                    if "__" in model_file.name:
                        continue
                    model_processor(model_file)

    def model(self, class_: str) -> DefaultMeta:
        """
        Returns the model class for the given ORM class name
        """
        return self.__model_registry__.class_(class_)

    def model_meta(self, class_: Union[str, DefaultMeta]) -> dict:
        """
        Returns meta information for the given ORM class name
        """

        def check_for_table_name(model_):
            if not hasattr(model_, "__tablename__"):
                raise AttributeError(f"{model_} is not a valid model")

        if isinstance(class_, str):
            model = self.__model_registry__.class_(class_)
            check_for_table_name(model)
            return {
                "location": model.__module__,
                "table_name": model.__tablename__,
            }

        return {
            "location": class_.__module__,
            "table_name": class_.__tablename__,
        }

    @deprecated("import_structures() will be removed, Use import_themes() instead")
    def import_structures(self, structures_folder: str) -> None:
        self.import_themes(structures_folder)

    @deprecated("model_class() will be removed, Use model() instead")
    def model_class(self, class_name: str) -> Any:
        return self.model(class_name)



class Resources:
    # Format to: NONE
    blueprint_init = """\
from flask_bigapp import Blueprint

bp = Blueprint(__name__)

bp.import_routes("routes")


@bp.before_app_request
def before_app_request():
    bp.init_session()

"""

    # Format to: name, static_url_path, url_prefix
    blueprint_config = """\
enabled = "yes"

[settings]
url_prefix = "{url_prefix}"
subdomain = ""
url_defaults = {{}}
static_folder = "static"
template_folder = "templates"
static_url_path = "{static_url_path}"

[session]
{name}_session = "yes"

"""

    # Format to: NONE
    blueprint_index_route = """\
from flask import render_template

from .. import bp


@bp.route("/", methods=["GET"])
def index():
    return render_template(bp.tmpl("index.html"))

"""

    # Format to: name
    blueprint_index_template = """\
<p>{name} index</p>
"""

    # Format to: name, static_url_path, url_prefix
    theme_config = """\
enabled = "yes"
static_folder = "static"
static_url_path = "{static_url_path}"
    """



def loader(app):
    @app.context_processor
    def example__utility_processor():
        """
        Usage

        {{ format_price(100) }}
        {{ format_price(100.33) }}
        """

        def example__format_price(amount, currency='$'):
            return '{1}{0:.2f}'.format(amount, currency)

        return dict(format_price=example__format_price)



import typing as t
from types import ModuleType

from flask_sqlalchemy.model import DefaultMeta


class ModelRegistry:
    registry: t.Dict[str, t.Any]

    def __init__(self):
        self.registry = dict()

    def assert_exists(self, class_name: str):
        if class_name not in self.registry:
            raise KeyError(
                f"Model {class_name} not found in model registry \n"
                f"Available models: {', '.join(self.registry.keys())}"
            )

    def add(self, ref, model: t.Optional[ModuleType] = None):
        self.registry[ref] = model

    def class_(self, class_name: str) -> DefaultMeta:
        self.assert_exists(class_name)
        return self.registry[class_name]

    def __repr__(self):
        return f"ModelRegistry({self.registry})"



import re
from dataclasses import dataclass
from datetime import datetime
from hashlib import sha1, sha256, sha512
from random import choice
from random import randrange
from string import punctuation, ascii_letters


class Auth:

    @classmethod
    def is_email_address_valid(cls, email_address: str) -> bool:
        """
        Checks if email_address is a valid email address.
        :param email_address:
        :return bool:
        """
        pattern = re.compile(
            r"[a-z\d!#$%&'*+/=?^_`{|}~-]+(?:\.[a-z\d!#$%&'*+/=?^_`{|}~-]+)*@(?:[a-z\d](?:[a-z\d-]*[a-z\d])?\.)+[a-z\d](?:[a-z\d-]*[a-z\d])?",
            re.IGNORECASE
        )
        return bool(pattern.match(email_address))

    @classmethod
    def is_username_valid(cls, username: str, allowed: str = "all") -> bool:
        """
        Allowed options: "all", "dot", "dash", "under"
        Checks if a username is valid. Valid usernames can only include
        letters, numbers, ., -, and _ but can not begin or end with the last three mentioned
        :param allowed:
        :param username:
        :return bool:
        """
        matches = list()
        if username.isalnum():
            return True
        if re.findall(r"([a-zA-Z\d]+(\.+[a-zA-Z\d]+)+)", username):
            matches.append("dot")
        if re.findall(r"([a-zA-Z\d]+(-[a-zA-Z\d]+)+)", username):
            matches.append("dash")
        if re.findall(r"([a-zA-Z\d]+(_[a-zA-Z\d]+)+)", username):
            matches.append("under")
        if allowed == "all" and matches:
            return True
        if allowed == "dot" and matches:
            return False
        if allowed == "dash" and matches:
            return False
        if allowed == "under" and matches:
            return False
        return False

    @classmethod
    def generate_form_token(cls) -> str:
        """
        Generates a SHA1 using today's date and time.
        :return str: hash:
        """
        sha = sha1()
        sha.update(str(datetime.now()).encode("utf-8"))
        return sha.hexdigest()

    @classmethod
    def generate_salt(cls) -> str:
        """
        Generates a string of 4 special characters, for use in password salting
        :return str:
        """
        return "".join(choice(punctuation) for _ in range(4))

    @classmethod
    def generate_private_key(cls, hook: str) -> str:
        """
        Generates a private key for api access from a passed in hook value.
        :return str: hash:
        """
        sha = sha256()
        sha.update(hook.encode("utf-8"))
        return sha.hexdigest()

    @classmethod
    def generate_numeric_validator(cls, length: int) -> int:
        """
        Generates (length) of random numbers.
        :return str:
        """
        start = int("1" * length)
        end = int("9" * length)
        return randrange(start, end)

    @classmethod
    def generate_email_validator(cls) -> str:
        """
        Generates a string of 8 random numbers, for use in MFA email
        :return str:
        """
        return str(cls.generate_numeric_validator(length=8))

    @classmethod
    def generate_pepper(cls, password: str):
        """
        Chooses a random letter from ascii_letters and joins it onto the user's password,
        this is used to pepper the password
        :param password:
        :return str:
        """
        return "".join(choice(ascii_letters) for _ in range(1)) + password

    @classmethod
    def sha_password(cls, password: str, salt: str, encrypt: int = 512) -> str:
        """
        Takes user's password, peppers in, salts it, then converts it to sha
        Can set encryption to 256/512 - 256 is system
        :param password:
        :param salt:
        :param encrypt:
        :return str: hash:
        """
        sha = sha512() if encrypt == 512 else sha256()
        sha.update((cls.generate_pepper(password) + salt).encode("utf-8"))
        return sha.hexdigest()

    @classmethod
    def auth_password(cls, input_password: str, database_password: str, database_salt: str,
                      encrypt: int = 512) -> bool:
        """
        Takes user's password (input_password), loops over all possible ascii_letters joining
        to the password as a pepper, then salts using salt value in the database, then converts it
        to sha, then compares that loop result to the database password to find a match
        Can set encryption to 256/512 - 256 is system
        :param input_password: str
        :param database_password: str
        :param database_salt: str
        :param encrypt: int
        :return bool:
        """
        for letter in list(ascii_letters):
            sha = sha512() if encrypt == 512 else sha256()
            sha.update((letter + input_password + database_salt).encode("utf-8"))
            if sha.hexdigest() == database_password:
                return True
        return False

    @classmethod
    def generate_password(cls, style: str = "animals", length: int = 3) -> str:
        """
        style options: "animals"

        Will return a plain text password based on choice of style and length.

        Combinations available:

        style: str "animals", length: int <number of animals returned>

        :param style: str
        :param length: int
        :return: str:
        """
        if style == "animals":
            final = []
            for i in range(length):
                random_index = randrange(0, len(PasswordGeneration.animals))
                final.append(PasswordGeneration.animals[random_index])
            return '-'.join(final)
        raise ValueError(f"Invalid style passed in {style}")


@dataclass
class PasswordGeneration:
    """
    This is a bank of words used to generate random passwords.
    """
    animals = ['Canidae', 'Felidae', 'Cat', 'Cattle', 'Dog', 'Donkey', 'Goat', 'Horse', 'Pig', 'Rabbit',
               'Aardvark', 'Aardwolf', 'Albatross', 'Alligator', 'Alpaca', 'Amphibian', 'Anaconda',
               'Angelfish', 'Anglerfish', 'Ant', 'Anteater', 'Antelope', 'Antlion', 'Ape', 'Aphid',
               'Armadillo', 'Asp', 'Baboon', 'Badger', 'Bandicoot', 'Barnacle', 'Barracuda', 'Basilisk',
               'Bass', 'Bat', 'Bear', 'Beaver', 'Bedbug', 'Bee', 'Beetle', 'Bird', 'Bison', 'Blackbird',
               'Boa', 'Boar', 'Bobcat', 'Bobolink', 'Bonobo', 'Bovid', 'Bug', 'Butterfly', 'Buzzard',
               'Camel', 'Canid', 'Capybara', 'Cardinal', 'Caribou', 'Carp', 'Cat', 'Catshark',
               'Caterpillar', 'Catfish', 'Cattle', 'Centipede', 'Cephalopod', 'Chameleon', 'Cheetah',
               'Chickadee', 'Chicken', 'Chimpanzee', 'Chinchilla', 'Chipmunk', 'Clam', 'Clownfish',
               'Cobra', 'Cockroach', 'Cod', 'Condor', 'Constrictor', 'Coral', 'Cougar', 'Cow', 'Coyote',
               'Crab', 'Crane', 'Crawdad', 'Crayfish', 'Cricket', 'Crocodile', 'Crow', 'Cuckoo', 'Cicada',
               'Damselfly', 'Deer', 'Dingo', 'Dinosaur', 'Dog', 'Dolphin', 'Donkey', 'Dormouse', 'Dove',
               'Dragonfly', 'Dragon', 'Duck', 'Eagle', 'Earthworm', 'Earwig', 'Echidna', 'Eel', 'Egret',
               'Elephant', 'Elk', 'Emu', 'Ermine', 'Falcon', 'Ferret', 'Finch', 'Firefly', 'Fish',
               'Flamingo', 'Flea', 'Fly', 'Flyingfish', 'Fowl', 'Fox', 'Frog', 'Gamefowl', 'Galliform',
               'Gazelle', 'Gecko', 'Gerbil', 'Gibbon', 'Giraffe', 'Goat', 'Goldfish', 'Goose', 'Gopher',
               'Gorilla', 'Grasshopper', 'Grouse', 'Guan', 'Guanaco', 'Guineafowl', 'Gull', 'Guppy',
               'Haddock', 'Halibut', 'Hamster', 'Hare', 'Harrier', 'Hawk', 'Hedgehog', 'Heron', 'Herring',
               'Hippopotamus', 'Hookworm', 'Hornet', 'Horse', 'Hoverfly', 'Hummingbird', 'Hyena', 'Iguana',
               'Impala', 'Jackal', 'Jaguar', 'Jay', 'Jellyfish', 'Junglefowl', 'Kangaroo', 'Kingfisher',
               'Kite', 'Kiwi', 'Koala', 'Koi', 'Krill', 'Ladybug', 'Lamprey', 'Landfowl', 'Lark', 'Leech',
               'Lemming', 'Lemur', 'Leopard', 'Leopon', 'Limpet', 'Lion', 'Lizard', 'Llama', 'Lobster',
               'Locust', 'Loon', 'Louse', 'Lungfish', 'Lynx', 'Macaw', 'Mackerel', 'Magpie', 'Mammal',
               'Manatee', 'Mandrill', 'Marlin', 'Marmoset', 'Marmot', 'Marsupial', 'Marten', 'Mastodon',
               'Meadowlark', 'Meerkat', 'Mink', 'Minnow', 'Mite', 'Mockingbird', 'Mole', 'Mollusk',
               'Mongoose', 'Monkey', 'Moose', 'Mosquito', 'Moth', 'Mouse', 'Mule', 'Muskox', 'Narwhal',
               'Newt', 'Nightingale', 'Ocelot', 'Octopus', 'Opossum', 'Orangutan', 'Orca', 'Ostrich',
               'Otter', 'Owl', 'Ox', 'Panda', 'Panther', 'Parakeet', 'Parrot', 'Parrotfish', 'Partridge',
               'Peacock', 'Peafowl', 'Pelican', 'Penguin', 'Perch', 'Pheasant', 'Pig', 'Pigeon', 'Pike',
               'Pinniped', 'Piranha', 'Planarian', 'Platypus', 'Pony', 'Porcupine', 'Porpoise', 'Possum',
               'Prawn', 'Primate', 'Ptarmigan', 'Puffin', 'Puma', 'Python', 'Quail', 'Quelea', 'Quokka',
               'Rabbit', 'Raccoon', 'Rat', 'Rattlesnake', 'Raven', 'Reindeer', 'Reptile', 'Rhinoceros',
               'Roadrunner', 'Rodent', 'Rook', 'Rooster', 'Roundworm', 'Sailfish', 'Salamander', 'Salmon',
               'Sawfish', 'Scallop', 'Scorpion', 'Seahorse', 'Shark', 'Sheep', 'Shrew', 'Shrimp',
               'Silkworm', 'Silverfish', 'Skink', 'Skunk', 'Sloth', 'Slug', 'Smelt', 'Snail', 'Snake',
               'Snipe', 'Sole', 'Sparrow', 'Spider', 'Spoonbill', 'Squid', 'Squirrel', 'Starfish',
               'Stingray', 'Stoat', 'Stork', 'Sturgeon', 'Swallow', 'Swan', 'Swift', 'Swordfish',
               'Swordtail', 'Tahr', 'Takin', 'Tapir', 'Tarantula', 'Tarsier', 'Termite', 'Tern', 'Thrush',
               'Tick', 'Tiger', 'Tiglon', 'Toad', 'Tortoise', 'Toucan', 'Trout', 'Tuna', 'Turkey',
               'Turtle', 'Tyrannosaurus', 'Urial', 'Vicuna', 'Viper', 'Vole', 'Vulture', 'Wallaby',
               'Walrus', 'Wasp', 'Warbler', 'Weasel', 'Whale', 'Whippet', 'Whitefish', 'Wildcat',
               'Wildebeest', 'Wildfowl', 'Wolf', 'Wolverine', 'Wombat', 'Woodpecker', 'Worm', 'Wren',
               'Xerinae', 'Yak', 'Zebra', 'Alpaca', 'Cat', 'Cattle', 'Chicken', 'Dog', 'Donkey', 'Ferret',
               'Gayal', 'Goldfish', 'Guppy', 'Horse', 'Koi', 'Llama', 'Sheep', 'Yak']



import logging
import os
from pathlib import Path

from toml import load as toml_load

from .resources import Resources
from .utilities import if_env_replace, capitalize_dict_keys, cast_to_bool, lower_dict_keys


def build_database_uri(database_config_value: dict, app) -> str:
    """
    Puts together the correct database URI depending on the type specified.

    Fails if type is not supported.
    """
    app_root = Path(app.root_path)

    db_type = if_env_replace(database_config_value.get("TYPE", "None"))
    db_name = if_env_replace(database_config_value.get('DATABASE_NAME', 'database'))

    db_location = if_env_replace(database_config_value.get("LOCATION", "db"))
    db_port = if_env_replace(str(database_config_value.get('PORT', 'None')))

    db_username = if_env_replace(database_config_value.get('USERNAME', 'None'))
    db_password = if_env_replace(database_config_value.get('PASSWORD', 'None'))

    db_allowed = ('postgresql', 'mysql', 'oracle')

    if db_type == "sqlite":
        if db_location is not None:

            if Path(db_location).exists():
                database = Path(db_location / f"{db_name}.db")
                return f"sqlite:///{database}"

            db_location_path = Path(app_root / db_location)
            db_location_path.mkdir(parents=True, exist_ok=True)
            db_location_file_path = db_location_path / f"{db_name}.db"
            return f"sqlite:///{db_location_file_path}"

        db_at_root = Path(app_root / f"{db_name}.db")
        return f"{db_type}:///{db_at_root}"

    if db_type in db_allowed:
        return f"{db_type}://{db_username}:{db_password}@{db_location}:{db_port}/{db_name}"

    raise ValueError(
        "Unknown database type, must be: postgresql / mysql / oracle / sqlite")


def init_app_config(config_file_path: Path, app) -> dict:
    """
    Processes the values from the configuration from_file.
    """
    if not config_file_path.exists():
        logging.critical("Config file was not found, creating default.config.toml to use")

        config_file_path.write_text(
            Resources.default_config.format(
                secret_key=os.urandom(24).hex())
        )

    config_suffix = ('.toml', '.tml')

    if config_file_path.suffix not in config_suffix:
        raise TypeError("Config from_file must be one of the following types: .toml / .tml")

    config = capitalize_dict_keys(toml_load(config_file_path))
    flask_config = capitalize_dict_keys(config.get("FLASK"))
    session_config = config.get("SESSION")
    database_config = capitalize_dict_keys(config.get("DATABASE"))

    if flask_config is not None and isinstance(flask_config, dict):
        for flask_config_key, flask_config_value in flask_config.items():
            app.config.update({str(flask_config_key): if_env_replace(flask_config_value)})

    if database_config is not None and isinstance(database_config, dict):
        app.config['SQLALCHEMY_BINDS'] = dict()
        for database_config_key, database_config_values in database_config.items():
            values = capitalize_dict_keys(database_config_values)
            if values.get("ENABLED", False):
                if database_config_key == "MAIN":
                    app.config['SQLALCHEMY_DATABASE_URI'] = f"{build_database_uri(values, app)}"
                    continue

                app.config['SQLALCHEMY_BINDS'].update({
                    str(database_config_key).lower(): f"{build_database_uri(values, app)}"
                })

    return {"FLASK": flask_config, "SESSION": session_config, "DATABASE": database_config}


def init_bp_config(blueprint_name: str, config_file_path: Path) -> tuple:
    """
    Attempts to load the and process the configuration from_file.
    """

    if not config_file_path.exists():
        raise FileNotFoundError(f"{blueprint_name} Blueprint config from_file {config_file_path.name} was not found")

    config_suffix = ('.toml', '.tml')

    if config_file_path.suffix not in config_suffix:
        raise TypeError("Config from_file must be one of the following types: .toml / .tml")

    config = capitalize_dict_keys(toml_load(config_file_path))

    enabled = cast_to_bool(config.get('ENABLED', False))

    if not enabled:
        return enabled, {}, {}

    session = config.get('SESSION', {})
    settings = lower_dict_keys(config.get('SETTINGS', {}))

    kwargs = dict()

    valid_settings = (
        'url_prefix', 'subdomain', 'url_defaults', 'static_folder', 'template_folder', 'static_url_path', 'root_path'
    )

    for setting in valid_settings:
        if setting == 'url_prefix':
            kwargs.update(
                {'url_prefix': settings.get('url_prefix') if settings.get('url_prefix') != "" else f"/{blueprint_name}"}
            )
            continue
        if setting in settings:
            if settings.get(setting, False):
                kwargs.update({setting: settings.get(setting)})

    return enabled, session, settings



import typing as t

from sqlalchemy import select, update, insert, delete, desc, asc  # type: ignore
from sqlalchemy.orm import Session  # type: ignore


class CrudMixin:
    __id_field__: str
    __session__: Session

    @classmethod
    def create(
            cls,
            values: t.Optional[t.Dict[str, t.Any]] = None,
            batch: t.Optional[t.List[t.Dict[str, t.Any]]] = None,
            wash_attributes: bool = False,
            ignore_attributes: t.Optional[t.List[str]] = None,
    ):
        if batch is not None:
            new_batch = []
            if wash_attributes:
                for item in batch:
                    temp_item = {**item}
                    for key, value in item.items():
                        if ignore_attributes:
                            if key in ignore_attributes:
                                del temp_item[key]
                                continue
                        if not hasattr(cls, key):
                            del temp_item[key]
                    new_batch.append(temp_item)

            process_batch = batch if not wash_attributes else new_batch
            result = cls.__session__.scalars(
                insert(cls).returning(cls),
                process_batch
            )
            cls.__session__.commit()
            return result.all()

        if values:
            temp_values = {**values}
        else:
            temp_values = {}

        if wash_attributes and values is not None:
            for key, value in values.items():
                if not hasattr(cls, key):
                    del temp_values[key]

        if ignore_attributes is not None and values is not None:
            for key in ignore_attributes:
                if key in values:
                    del temp_values[key]

        result = cls.__session__.scalar(
            insert(cls).returning(cls).values(**temp_values)  # type: ignore
        )
        cls.__session__.commit()
        return result

    @classmethod
    def read(
            cls,
            id_: t.Optional[int] = None,
            field: t.Optional[tuple] = None,
            fields: t.Optional[t.Dict[str, t.Any]] = None,
            all_rows: bool = False,
            order_by: t.Optional[str] = None,
            order_desc: bool = False,
            _updating: bool = False,
            _deleting: bool = False,
            _auto_output: bool = True
    ):
        if _updating:
            base_query = update(cls)  # type: ignore
        elif _deleting:
            base_query = delete(cls)  # type: ignore
        else:
            if order_by is not None:
                column = getattr(cls, order_by)
                if order_desc:
                    base_query = select(cls).order_by(desc(column))  # type: ignore
                else:
                    base_query = select(cls).order_by(asc(column))  # type: ignore
            else:
                base_query = select(cls)  # type: ignore

        if all_rows:
            return cls.__session__.scalars(base_query).all()

        if id_ is not None:
            kwargs = {cls.__id_field__: id_}
            base_query = base_query.filter_by(**kwargs)
            if _updating or _deleting:
                return base_query
            if _auto_output:
                return cls.__session__.scalars(base_query).one_or_none()
            return cls.__session__.scalars(base_query)

        if field is not None:
            base_query = base_query.filter_by(**{field[0]: field[1]})
            if _updating or _deleting:
                return base_query
            if _auto_output:
                return cls.__session__.scalars(base_query).all()
            return cls.__session__.scalars(base_query)

        if fields is not None:
            base_query = base_query.filter_by(**fields)
            if _updating or _deleting:
                return base_query
            if _auto_output:
                return cls.__session__.scalars(base_query).all()
            return cls.__session__.scalars(base_query)

    @classmethod
    def update(
            cls,
            values: dict,
            id_: t.Optional[int] = None,
            field: t.Optional[tuple] = None,
            fields: t.Optional[t.Dict[str, t.Any]] = None,
            return_updated: bool = False,
            wash_attributes: bool = False,
            ignore_attributes: t.Optional[t.List[str]] = None,
    ):
        if values:
            temp_values = {**values}
        else:
            temp_values = {}

        if wash_attributes:
            for key, value in values.items():
                if not hasattr(cls, key):
                    del temp_values[key]

        if ignore_attributes is not None and values is not None:
            for key in ignore_attributes:
                if key in values:
                    del temp_values[key]

        query = cls.read(id_=id_, field=field, fields=fields, _updating=True).values(temp_values)
        if return_updated:
            result = cls.__session__.execute(query.returning(cls)).scalars().all()
            cls.__session__.commit()
            return result
        cls.__session__.execute(query)
        cls.__session__.commit()
        return True

    @classmethod
    def delete(
            cls,
            id_: t.Optional[int] = None,
            field: t.Optional[tuple] = None,
            fields: t.Optional[t.Dict[str, t.Any]] = None,
            return_deleted: bool = False
    ):
        query = cls.read(id_=id_, field=field, fields=fields, _deleting=True)
        if return_deleted:
            result = cls.__session__.execute(query.returning(cls)).scalars().all()
            cls.__session__.commit()
            return result[0] if len(result) > 0 else result
        cls.__session__.execute(query)
        cls.__session__.commit()
        return True
